from string import *

import urwid

class EmacsEdit(urwid.Edit):
    "An edit widget which handles basic standard emacs-style editing keys"
    def __init__(self, caption = "", edit_text = "", multiline = False,
		    align = 'left', wrap = 'space', allow_tab = False,
		    edit_pos = None, layout=None, edit_mode="insert",
		    read_only=False):
	self.edit_mode = edit_mode
	self.read_only = read_only
	urwid.Edit.__init__(self, caption, edit_text, multiline, align, wrap,
		allow_tab, edit_pos, layout)
    def keypress(self, (maxcol,), key):
	translationDict = {
		    "ctrl f": "right",
		    "ctrl b": "left",
		    "ctrl n": "down",
		    "ctrl p": "up",
		    "ctrl h": "backspace",
		    "ctrl d": "delete",
		    "ctrl a": "home",
		    "ctrl e": "end"
		}
	key = translationDict.get(key, key)

	if not self.read_only:
	    if self.edit_mode == 'replace' and self.valid_char(key):
		self._delete_highlighted()
		if type(key) == type(u""):
			key = key.encode("utf-8")
		self.replace_text(key)
		return
	    elif key in ["ctrl u", "ctrl k"]:
		self._delete_highlighted()
		self.pref_col_maxcol = None, None
		p = self.edit_pos
		x,y = self.get_cursor_coords((maxcol,))
		if key == "ctrl u":
		    self.move_cursor_to_coords((maxcol,),'left',y)
		    if self.edit_pos == p and self.edit_pos > 0:
			self.set_edit_pos(self.edit_pos - 1)
		    self.deleteFromTo(self.edit_pos, p)
		elif key == "ctrl k":
		    self.move_cursor_to_coords((maxcol,),'right',y)
		    if self.edit_pos == p and self.edit_pos < len(self.edit_text)-1:
			self.set_edit_pos(self.edit_pos + 1)
		    self.deleteFromTo(p, self.edit_pos)
		    self.set_edit_pos(p)
		return
	    elif key == "ctrl w":
		self._delete_highlighted()
		self.pref_col_maxcol = None, None
		p = self.edit_pos
		x,y = self.get_cursor_coords((maxcol,))

		self.toPrevWord()
		newx,newy = self.get_cursor_coords((maxcol,))
		if newy < y:
		    self.move_cursor_to_coords((maxcol,),'right',y-1)
		self.deleteFromTo(self.edit_pos, p)
		return

	if key == "meta f":
	    self.toNextWord()
	    return
	elif key == "meta b":
	    self.toPrevWord()
	    return
	elif key in ("page up","page down"):
	    # XXX complete hack, but better than nothing
	    self.highlight = None
	    
	    x,y = self.get_cursor_coords((maxcol,))
	    pref_col = self.get_pref_col((maxcol,))
	    assert pref_col is not None

	    if key == "page up": y -= 10
	    else:		y += 10

	    trans = self.get_line_translation(maxcol)
	    top_x, top_y = self.position_coords(maxcol, 0)
	    if y < top_y:
		y = top_y
	    if y >= len(trans):
		y = len(trans)-1

	    if not self.move_cursor_to_coords((maxcol,),pref_col,y):
		    return key

	if self.read_only:
	    if self.valid_char(key) or key in ["delete", "backspace"]:
		return

	return urwid.Edit.keypress(self, (maxcol,), key)

    def deleteFromTo(self, fr, to):
	self.set_edit_text( self.edit_text[:fr] +
		self.edit_text[to:] )
    def toNextWord(self):
	p = self.edit_pos+1
	l = len(self.edit_text)
	while p < l and not self.edit_text[p] in ' \n\t':
	    p+=1
	while p < l and self.edit_text[p] in ' \n\t':
	    p+=1
	self.set_edit_pos(p)
    def toPrevWord(self):
	p = self.edit_pos-1
	while p >= 0 and self.edit_text[p] in ' \n\t':
	    p-=1
	while p >= 0 and not self.edit_text[p] in ' \n\t':
	    p-=1
	self.set_edit_pos(p+1)

    def replace_text(self, text):
	    """Replace text at the cursor position"""
	    p = self.edit_pos
	    if p < len(self.edit_text) and self.edit_text[p] == '\n': text+='\n'
	    self.set_edit_text( self.edit_text[:p] + text +
		    self.edit_text[p+1:] )

