from string import *

from util import *
from BigTable import BigTable
from zeroSum import solveZeroSumMatrix

def escapeChars(s, chars):
    ret = ''
    for c in s:
	if c in '\\'+chars:
	    ret += '\\'
	ret += c
    return ret

def prettyRepr(val):
    if val is None: return "None"
    if type(val) == str:
	if val.find('\n') == -1:
	    return '"%s"' % escapeChars(val, '"')
	else:
	    return '"""%s"""' % escapeChars(val, '"')
    else:
	raise "unimplemented"

class Valuation:
    def __init__(self, valuation=''):
	self.set(valuation)
    def __str__(self):
	if self.valuation is None: return '--'
	return self.valuation
    def set(self, string):
	if string is '':
	    self.valuation = None
	    return
	self.valuation = string
    def prettyConstructorStr(self):
	return "Valuation(%s)" % prettyRepr(self.valuation)

class Choice:
    def __init__(self, history="", note="", mixed=None):
	self.history, self.note, self.mixed = history, note, mixed
    def flatten(self):
	ret = escapeChars(self.history, ':')
	if self.note != "":
	    ret += '\nN:%s' % escapeChars(self.note, ':')
	if self.mixed is not None:
	    ret += '\nM:%s' % self.mixed
	return ret
    def __str__(self):
	ret = self.history
	if self.mixed is not None: ret = "%s\n%s" % (self.mixed, ret)
	if self.note != "": ret += '\n%s' % self.note
	return ret
    def prettyConstructorStr(self):
	ret = "Choice("
	ret += prettyRepr(self.history)
	if self.note != "": ret += ",\nnote = %s" % prettyRepr(self.note)
	if self.mixed != None: ret += ",\nmixed = %s" % self.mixed
	ret += ')'
	return ret

class Node(object):
    def __init__(self, identifier, historyR=None, historyC=None, state=None,
	    notes=None, valuation=None,
	    parent=None, parentRow=None, parentCol=None,
	    rowChoices=None, colChoices=None, children=None):
	if valuation is None: valuation = Valuation()
	if rowChoices is None: rowChoices = []
	if colChoices is None: colChoices = []
	if children is None: children = []
	(self.identifier, self._historyR, self._historyC, self._state,
		self._notes, self.valuation, self.parent, self.parentRow,
		self.parentCol, self.rowChoices, self.colChoices,
		self.children) = (identifier, historyR, historyC, state,
			notes, valuation, parent, parentRow, parentCol,
			rowChoices, colChoices, children)

	self._childcount = 0
	self.matrix_invalidated = False

    def getHistoryR(self):
	if self._historyR is None:
	    if self.parent: return appendRotated(self.parent.historyR,
		    self.parent.rowChoices[self.parentRow].history)
	    else: return ""
	else: return self._historyR
    def setHistoryR(self, h):
	if h != self.historyR: self._historyR = h
    historyR = property(getHistoryR, setHistoryR)
    def getHistoryC(self):
	if self._historyC is None:
	    if self.parent: return appendRotated(self.parent.historyC,
		    self.parent.colChoices[self.parentCol].history)
	    else: return ""
	else: return self._historyC
    def setHistoryC(self, h):
	if h != self.historyC: self._historyC = h
    historyC = property(getHistoryC, setHistoryC)

    def getState(self):
	if self._state is None:
	    if self.parent: return self.parent.state
	    else: return ""
	else: return self._state
    def setState(self, s):
	if s != self.state: self._state = s
    state = property(getState, setState)

    def getNotes(self):
	if self._notes is None:
	    if self.parent:
		def noteIt(label, note):
		    if note:
			ret = ''
			firstLine = True
			for line in note.splitlines():
			    if firstLine:
				ret += label+': '
				firstLine = False
			    else: ret += '\t'
			    ret += line+'\n'
			return ret
		    else: return ''
		return (noteIt("Row Choice", self.parent.rowChoices[self.parentRow].note) +
			noteIt("Col Choice", self.parent.colChoices[self.parentCol].note))
	    else: return ""
	else: return self._notes
    def setNotes(self, s):
	if s != self.notes: self._notes = s
    notes = property(getNotes, setNotes)

    def getMatrixWidget(self, startRow=0, startCol=0, selectedRow=None, selectedCol=None):
	def attrOfCell(row, col):
	    if ((row >=0 and self.rowChoices[row].mixed == 0) or
		    (col >= 0 and self.colChoices[col].mixed == 0)):
		return "dominated"
	    else:
		return None
	return BigTable(str(self.valuation),
		map(str, self.colChoices),
		map(str, self.rowChoices),
		[ [ str(child.valuation) for child in row ] for row in self.children ],
		startRow, startCol, selectedRow, selectedCol,
		attrOfCell=attrOfCell)

    def _invalidate_matrix(self):
	self.matrix_invalidated = True

    def addRowChoice(self, choice):
	self.rowChoices.append(choice)
	self.children.append(
		[ Node('%s_%s' % (self.identifier, attrPlusplus(self, '_childcount')),
			parent = self,
			parentRow = len(self.rowChoices)-1,
			parentCol = i,
			) for i in range(len(self.colChoices)) ])
	self._invalidate_matrix()

    def addColChoice(self, choice):
	self.colChoices.append(choice)
	for i in range(len(self.children)):
	    self.children[i].append(
		    Node('%s_%s' % (self.identifier, attrPlusplus(self, '_childcount')),
			parent = self,
			parentRow = i,
			parentCol = len(self.colChoices)-1,
			))
	self._invalidate_matrix()

    def editRowChoice(self, row, history=None, note=None):
	if history is not None: self.rowChoices[row].history = history
	if note is not None: self.rowChoices[row].note = note
	self._invalidate_matrix()
    def editColChoice(self, col, history=None, note=None):
	if history is not None: self.colChoices[col].history = history
	if note is not None: self.colChoices[col].note = note
	self._invalidate_matrix()
    def deleteRowChoice(self, row):
	del self.rowChoices[row]
	del self.children[row]
	for childRow in self.children[row:]:
	    for child in childRow:
		child.parentRow -= 1
	self._invalidate_matrix()
    def deleteColChoice(self, col):
	del self.colChoices[col]
	for childRow in self.children:
	    del childRow[col]
	    for child in childRow[col:]:
		child.parentCol -= 1
	self._invalidate_matrix()
    def dominateRowChoice(self, row):
	if self.rowChoices[row].mixed == 0:
	    self.rowChoices[row].mixed = None
	else:
	    self.rowChoices[row].mixed = 0
	self._invalidate_matrix()
    def dominateColChoice(self, col):
	if self.colChoices[col].mixed == 0:
	    self.colChoices[col].mixed = None
	else:
	    self.colChoices[col].mixed = 0
	self._invalidate_matrix()

    def flatten(self):
	def multiline(string):
	    return join(["\n%" + line for line in string.splitlines()],'')
	def listString(strings):
	    if strings == []: return ""
	    return '[' + join(map(lambda s: escapeChars(s, ',]'),strings), ',') + ']'
	def listStringList(stringsList):
	    if stringsList == []: return ""
	    return '[' + join(map(listString, stringsList), ',') + ']'

	string = "Id: %s\n" % self.identifier
	if self.parent:
	    string += "Parent: %s\n" % self.parent.identifier
	    string += "ParentCoords: [%s,%s]\n" % (str(self.parentRow), str(self.parentCol))
	if self.valuation.valuation:
	    string += "Valuation: %s\n" % str(self.valuation.valuation)
	if self._historyR is not None:
	    string += "HistoryR:%s\n" % multiline(self._historyR)
	if self._historyC is not None:
	    string += "HistoryC:%s\n" % multiline(self._historyC)
	if self._state is not None:
	    string += "State:%s\n" % multiline(self._state)
	if self._notes is not None:
	    string += "Notes:%s\n" % multiline(self._notes)
	if self.rowChoices != []:
	    string += "RowChoices:%s\n" % listString([
		    multiline(choice.flatten()) for choice in
		    self.rowChoices])
	if self.colChoices != []:
	    string += "ColChoices:%s\n" % listString([
		    multiline(choice.flatten()) for choice in
		    self.colChoices])
	if self.children != []:
	    string += "Children:%s\n" % (
			listStringList([ [ child.identifier for child in row ]
				for row in self.children ]))
	string += "-----\n"

	return string

    def flattenTree(self):
	return self.flatten() + join([ child.flattenTree() for row in
		self.children for child in row ], '')

    def prettyConstructorStr(self):
	ret = "Node("
	ret += prettyRepr(self.identifier)
	if self._historyR is not None: ret += ",\n\thistoryR = %s" % \
	    prettyRepr(self._historyR)
	if self._historyC is not None: ret += ",\n\thistoryC = %s" % \
	    prettyRepr(self._historyC)
	if self._state is not None: ret += ",\n\tstate = %s" % \
	    prettyRepr(self._state)
	if self._notes is not None: ret += ",\n\tnotes = %s" % \
	    prettyRepr(self._notes)
	if self.valuation.valuation is not None:
	    ret += ",\n\tvaluation = %s" % \
			self.valuation.prettyConstructorStr()
	if self.rowChoices != []: ret += ",\n\trowChoices = [\n\t\t%s\n\t]" % \
	    join( [ c.prettyConstructorStr() for c in self.rowChoices], ',\n\t\t')
	if self.colChoices != []: ret += ",\n\tcolChoices = [\n\t\t%s\n\t]" % \
	    join( [ c.prettyConstructorStr() for c in self.colChoices], ',\n\t\t')
	if self.parent is not None: ret += ",\n\tparent = %s" % \
	    repr(self.parent.identifier)
	if self.parentRow is not None: ret += ",\n\tparentRow = %s" % \
	    str(self.parentRow)
	if self.parentCol is not None: ret += ",\n\tparentCol = %s" % \
	    str(self.parentCol)
	if self.children != []: ret += ",\n\tchildren=[%s]" % (
		join( [ "[%s]" % join( [ repr(child.identifier) for child in
			childRow ], ',') for childRow in self.children ]
		    , ','))
	ret += ')'
	return ret

    def pyTreeStr(self):
	return "registeredNodes = registerNodes([\n%s\n])\n" % self._pyTreeStr()
    def _pyTreeStr(self):
	return join( [self.prettyConstructorStr()] +
		    [ child._pyTreeStr()
			    for childRow in self.children
			    for child in childRow
			], ',\n' )

    def getRoot(self):
	node = self
	while node.parent:
	    node = node.parent
	return node

    def setValuation(self, newVal):
	self.valuation.set(newVal)
	self._invalidate_matrix()

    def setChildValuation(self, row, col, newVal):
	self.children[row][col].valuation.set(newVal)
	self._invalidate_matrix()

    def _solve(self):
	if len(self.rowChoices) == 0 or len(self.colChoices) == 0:
	    raise UserError("Evaluation failed: null matrix")
	try: zeroSumMatrix = [
		    [ float(child.valuation.valuation) for child in childRow ]
		    for childRow in self.children ]
	except (ValueError, TypeError):
	    raise UserError("Evaluation failed: children must have numeric valuations")

	valuation, colMixed = solveZeroSumMatrix(zeroSumMatrix)

	transpose = [ [ -row[i] for row in zeroSumMatrix ] for i in
	    range(len(zeroSumMatrix[0])) ]
	valuation2, rowMixed = solveZeroSumMatrix(transpose)
	assert valuation2 == -valuation

	return valuation, colMixed, rowMixed

    def evaluate(self):
	try:
	    valuation, colMixed, rowMixed = self._solve()
	except IOError:
	    raise UserError('evaluation failed - is package glpk installed?')
	# TODO maybe: store accurate values, round only for display
	self.setValuation(str(int(round(valuation))))
	for i,choice in enumerate(self.rowChoices):
	    choice.mixed = int(round(100*rowMixed[i]))
	for i,choice in enumerate(self.colChoices):
	    choice.mixed = int(round(100*colMixed[i]))
	self._invalidate_matrix()

    def plot(self, forCol=False):
	"returns list of commands to send to gnuplot"

	matrix = self.children

	if forCol and matrix != []:
	    # transpose
	    matrix = [
		[ matrix[row][col]
		    for row in range(len(matrix)) ]
		for col in range(len(matrix[0])) ]

	if not forCol:
	    undominatedRows = [
		n
		for n in range(len(self.rowChoices))
		if self.rowChoices[n].mixed != 0 ]
	else:
	    undominatedRows = [
		n
		for n in range(len(self.colChoices))
		if self.colChoices[n].mixed != 0 ]

	if len(undominatedRows) > 3:
	    undominatedRows = undominatedRows[0:3]

	if len(undominatedRows) < 2:
	    raise UserError("Insufficient choices to plot")

	try: constraints = [
		    [ float(matrix[row][col].valuation.valuation)
			for row in undominatedRows ]
		    for col in range(len(matrix[0])) ]
	except (ValueError, TypeError):
	    raise UserError("Plotting failed: children must have numeric valuations")
	
	if constraints == []:
	    raise UserError("No constraints!")

	def choiceLabel(choice):
	    if choice.note != "":
		str="%s; %s" % (choice.history, choice.note)
	    else:
		str = choice.history
	    return replace(str, "\n", "/")
	if not forCol:
	    rowNames = [ choiceLabel(self.rowChoices[i])
		for i in undominatedRows ]
	    colNames = [ choiceLabel(self.colChoices[i])
		for i in range(len(constraints)) ]
	else:
	    rowNames = [ choiceLabel(self.colChoices[i])
		for i in undominatedRows ]
	    colNames = [ choiceLabel(self.rowChoices[i])
		for i in range(len(constraints)) ]

	
	commands = []
	if len(undominatedRows) == 3:
	    commands.append('set xrange [0:1]')
	    commands.append('set yrange [0:1]')
	    commands.append('set xlabel "%s"' % rowNames[0])
	    commands.append('set ylabel "%s"' % rowNames[1])
	    commands.append('set view 45, 335')
	    graphcommand = "splot " + join(
		    [ 'x+y <= 1 ? (%f)*x+(%f)*y+(%f)*(1-x-y) : 1/0  title "%s"'
			% tuple(constraint + [colNames[i]])
			    for i,constraint in enumerate(constraints) ],
		    ',')
	    commands.append(graphcommand)
	else: 
	    assert len(undominatedRows) == 2
	    commands.append('set xrange [0:1]')
	    commands.append('set yrange restore')
	    commands.append('set xlabel "%s"' % rowNames[0])
	    graphcommand = "plot " + join(
		    [ '(%f)*x+(%f)*(1-x)  title "%s"'
			% tuple(constraint + [colNames[i]])
			    for i,constraint in enumerate(constraints) ],
		    ',')
	    commands.append(graphcommand)
	return commands

    def clearMixed(self):
	for choice in self.rowChoices: choice.mixed = None
	for choice in self.colChoices: choice.mixed = None
	self._invalidate_matrix()
    
