#!/usr/bin/env python

import os,sys
from string import *
import glob

#import urwid_svn as urwid
#import urwid_svn.curses_display as curses_display
import urwid
import urwid.raw_display as raw_display
import urwid.web_display as web_display

import snarfRavenblack

from util import *
from BigTable import BigTable
from Node import Node, Choice, Valuation
#from parse import unflattenTree
from parsePy import readPyTree
from EmacsEdit import EmacsEdit
from GnuplotPipe import GnuplotPipe


if web_display.is_web_request():
    Screen = web_display.Screen
    chDir = "./trees"
    chrooted = True
else:
    chDir = "./trees"
    chrooted = True
    try:
	import urwid.curses_display as curses_display
	Screen = curses_display.Screen
    except:
	Screen = raw_display.Screen

class EphEdit(EmacsEdit):
    """
    Specialised for use by ephEditWidget
    (i.e. EmacsEdit with added hacks)
    """
    def __init__(self, caption = "", edit_text = "", multiline = False,
		    align = 'left', wrap = 'space', allow_tab = False,
		    edit_pos = None, layout=None, edit_mode="insert",
		    get_completions=None):
	self.get_completions = get_completions
	EmacsEdit.__init__(self, caption, edit_text, multiline, align, wrap,
		allow_tab, edit_pos, layout, edit_mode)
    def keypress(self, (maxcol,), key):
	if self.allow_tab == "complete":
	    if key == "tab":
		if self.complete_list == None:
		    if self.edit_pos == 0 or self.edit_text[self.edit_pos-1] in " \n\t":
			comp = ""
			self.complete_from = self.edit_pos
		    else:
			p = self.edit_pos
			self.toPrevWord()
			comp = self.edit_text[self.edit_pos:p]
			self.deleteFromTo(self.edit_pos, p)
			self.complete_from = self.edit_pos
		    self.complete_list = self.get_completions(comp)
		    if not comp in self.complete_list: self.complete_list.append(comp)
		    self.complete_i = 0
		else:
		    self.deleteFromTo(self.complete_from, self.edit_pos)
		    self.set_edit_pos(self.complete_from)

		self.insert_text(self.complete_list[self.complete_i])
		self.complete_i = (self.complete_i + 1) % len(self.complete_list)

		return

	    else:
		self.complete_list = None

	if (key == "enter" and self.multiline is True) or (
		key == "ctrl o" and self.multiline == "command"):
	    self._delete_highlighted()
	    key = "\n"
	    self.insert_text( key )
	elif key == "enter":
	    # don't let Edit.keypress handle it
	    return key
	else:
	    return EmacsEdit.keypress(self, (maxcol,), key)

    def nextMatch(self, list, i, partial):
	for j in range(1, len(list)):
	    k = (i + j) % len(list)
	    if list[k][:len(partial)] == partial:
		return k, list[k]
	return i, partial

HELP_TEXT = """
Simultaneous Move Game Explorer
===============================

Keybindings:
        (Notation: '^A' = control-a; 'M-a' = meta-a = alt-a)

    ?: this help
    Arrow keys or ^N ^P ^F ^B: navigate matrix
    Enter, Delete: context-dependent
    u: up to parent position
    U: up to root position

    r: add row choice
    c: add col choice
    R: edit selected row choice
    C: edit selected col choice

    ^R: load position tree (losing current tree!)
    ^S: save position tree
    M-T: start new tree (losing current tree!)
    q: quit

    H: edit position history (top is row history, bottom is col history)
    S: edit position state
    N: edit position notes

    v: set valuation of current position
    V: set valuation of currently selected child position
    D: mark current row/col as dominated
    e: evaluate current position (requires glpsol)
    M: remove mixed strategy and domination annotations

    p,P: plot current matrix (requires gnuplot)

    M-r: edit selected row note
    M-c: edit selected col note
    M-R: delete selected row
    M-C: delete selected col

Hints for use:
    SMGE is a dumb tool. It does not understand Waving Hands, nor any other game you may use it for. It is currently largely undocumented - the best way to get a feel for its uses is to play around with it.

    As there is no One True best concise notation system for Waving Hands, SMGE leaves most choices on how to represent moves and positions up to the user. Here are a few hints on how to do that.

    The choices (as entered with 'r' and 'c') should indicate as much of the pertinent information as can be expressed in a few characters. The choice text is rotated and added to the history to form the history of the next position - as indicated by the red preview text when editing the choice text.
    
    For example, a choice which involves waving D on the left hand and P on the right hand, completing DPP on the right hand and aiming it at the caster, might be represented concisely (losing information) by ' DP<' (with the '<' indicating the target).

    If further information is necessary (e.g. release of a banked spell, or choice of gesture to charm opponent to, or such), it may be better to add this as a choice note (M-r and M-c). For example, if the caster in the previous example chose to cast shield rather than DPP, this could be indicated in a note. Such do not appear explicitly in the new history, though note that you can edit the history of a position at any time with 'H'.
"""

class QuitException(Exception):
    pass

class Interface:
    def __init__(self, initialNode, initialFilename=None):
	self.currentFilename = initialFilename

	self.historyRWidget = EmacsEdit('', '',
		    multiline=True, wrap='clip', allow_tab = True,
		    align = 'right')
	self.historyCWidget = EmacsEdit('', '',
		    multiline=True, wrap='clip', allow_tab = True,
		    align = 'right')
	self.historyRPreviewWidget = urwid.Text('')
	self.historyCPreviewWidget = urwid.Text('')
	self.historyRPreviewWidgetWrap = urwid.AttrWrap(
		self.historyRPreviewWidget, 'history_preview')
	self.historyCPreviewWidgetWrap = urwid.AttrWrap(
		self.historyCPreviewWidget, 'history_preview')
	self.historyWidgetWrap = urwid.Pile( [
		urwid.Columns([self.historyRWidget, self.historyRPreviewWidgetWrap]),
		urwid.Divider('-'),
		urwid.Columns([self.historyCWidget, self.historyCPreviewWidgetWrap])
	    ] )

	self.stateWidget = EmacsEdit('State\n', '', multiline=True,
		    wrap='space', allow_tab = True)
	#self.stateWidgetWrap = urwid.Filler( self.stateWidget, valign='top')
	self.stateWidgetWrap = self.stateWidget

	self.notesWidget = EmacsEdit('Notes\n', '', multiline=True,
		    wrap='space', allow_tab = True)
	#self.notesWidgetWrap = urwid.Filler( self.notesWidget, valign='top')
	self.notesWidgetWrap = self.notesWidget

	self.messageWidget = urwid.Text('')
	self.ephEditWidget = EphEdit('','', wrap='clip', allow_tab = True)
	self.matrixWidgetWrap = urwid.WidgetWrap(urwid.Text(""))

	# TODO: proper scrolling text widget - ought to be able to use ListBox
	#   for this? Doesn't seem to work...
	#self.helpTextWidget = urwid.Text(HELP_TEXT)
	#self.helpTextWidgetWrapper = urwid.ListBox([self.helpTextWidget])
	
	# XXX: for now, we hack it with a read-only edit box
	self.helpTextWidget = EmacsEdit('',HELP_TEXT, edit_pos=0, read_only=True)
	self.helpTextWidgetWrapper = urwid.LineBox(urwid.Filler(self.helpTextWidget))

	self.setCurrentNode(initialNode)
	self.upperList = [ self.historyWidgetWrap, self.stateWidget,
		    self.notesWidget ]
	self.setGeometry()

	self.inputState = 'top'
	self.top.set_focus(self.lower)

	self.gnuplotPipe = GnuplotPipe()

    def setWidgetContents(self):
	self.historyRWidget.set_edit_text(self.currentNode.historyR)
	self.historyCWidget.set_edit_text(self.currentNode.historyC)
	self.stateWidget.set_edit_text(self.currentNode.state)
	self.notesWidget.set_edit_text(self.currentNode.notes)
	self.matrixWidget = self.currentNode.getMatrixWidget()
	self.matrixWidgetWrap.w = self.matrixWidget

    def setGeometry(self):
	self.upper = urwid.Columns( intersperse(
	    self.upperList,
	    ('fixed', 1, urwid.SolidFill('|'))),
	    box_columns=[2*i+1 for i in range(len(self.upperList)-1)])
	self.lower = self.matrixWidgetWrap
	self.top = urwid.Pile( [
		('flow', self.upper),
		('fixed', 1, urwid.SolidFill('-')),
		('weight', 2, self.lower),
		('fixed', 1, urwid.SolidFill('-')),
		('flow', self.messageWidget),
		('flow', self.ephEditWidget),
	    ] )

    def selectUpperWidget(self, widget):
	if not widget in self.upperList:

	    self.upperList.append(widget)
	    # TODO: sorting

	    self.setGeometry()

	self.upper.set_focus(widget)

    def removeUpperWidget(self, widget):
	if widget in self.upperList and len(self.upperList) > 1:
	    self.upperList.remove(widget)
	    self.setGeometry()

    def focusOnMatrix(self):
	self.top.set_focus(self.lower)

    def main(self):
	self.ui = Screen()
	self.ui.register_palette([
		('selected', 'light green', 'default', 'bold'),
		('dominated', 'dark red', 'default', 'standout'),
		('history_preview', 'light red', 'default', 'bold'),
		('confirm', 'white', 'default', 'standout'),
		('new-info', 'white', 'default', 'standout'),
		('old-info', 'default', 'default', None),
		('error', 'white', 'dark red', 'standout'),
	    ])
	self.ui.run_wrapper(self.run)

    def run(self):
	size = self.ui.get_cols_rows()

	try:
	    while True:
		if self.currentNode.matrix_invalidated:
		    self.resetMatrix()
		self.draw_screen(size)
		try:
		    keys = self.ui.get_input()
		except KeyboardInterrupt:
		    if self.inputState in ["top", "waitKey"]:
			keys = ['q']
		    else:
			keys = ['esc']
		for k in keys:
		    if k == "window resize":
			size = self.ui.get_cols_rows()
			continue
		    else:
			self.keypress(size, k)
	except QuitException:
	    return

    def keypress(self, size, k):
	node = self.currentNode

	if self.inputState == 'waitKey':
	    self.inputState = 'top'
	    self.focusOnMatrix()
	    if self.waitKeyCallback: k = self.waitKeyCallback(k)
	    if not k: return
	
	# fallthrough
	try:
	    if self.inputState == 'top':
		if k == 'q':
		    self.confirmAction('Quit?',
			    lambda: self.quit())
		elif k == '?':
		    self.overlay(self.helpTextWidgetWrapper)
		elif k in 'HSN':
		    widget,name = {
			'H': (self.historyWidgetWrap,'history'),
			'S': (self.stateWidgetWrap,'state'),
			'N': (self.notesWidgetWrap,'notes'),
			}[k]
		    self.initUpperEdit(widget,
			    "Editing %s; <esc> to finish" % name)
		elif k in ['meta H', 'meta S', 'meta N']:
		    self.removeUpperWidget(
			    {
				    'meta H': self.historyWidgetWrap,
				    'meta S': self.stateWidgetWrap,
				    'meta N': self.notesWidgetWrap,
				}[k]
			    )
		    self.focusOnMatrix()
		elif k == 'u':
		    parent = node.parent
		    if parent:
			self.setCurrentNode(parent)

			# Select the cell of the child we're coming from.
			# FIXME: doesn't set startRow,startCol. Really
			# BigTable should have a method for that.
			for (i,childRow) in enumerate(parent.children):
			    for (j,child) in enumerate(childRow):
				if child == node:
				    self.matrixWidget.selectedRow = i
				    self.matrixWidget.selectedCol = j
		elif k == 'U':
		    self.setCurrentNode(node.getRoot())
		elif k in ['ctrl s', '}']:
		    self.initEphEdit('Save tree to file: [ Tab: complete ]', self.saveTree,
			    get_completions=lambda f: glob.glob('%s*'%f),
			    initialText = self.currentFilename)
		elif k in ['ctrl r', '{']:
		    self.initEphEdit('Load tree from file: [ Tab: complete ]', self.loadTree,
			get_completions=lambda f: glob.glob('%s*'%f))
		elif k == 'meta W':
		    self.initEphEdit('Import tree from ravenblacks - enter game number or warlock name:', self.warlocksTree)
		elif k == 'meta T':
		    self.confirmAction('Erase tree and start a new one?',
			    lambda: self.initEphEdit('Name for new tree:',
				lambda s: self.setCurrentNode(Node(s))))
		else:
		    k = self.top.keypress(size, k)
		    row = self.matrixWidget.selectedRow
		    col = self.matrixWidget.selectedCol

		    if k == 'r':
			self.initEphEdit('New row choice: [ ^O: newline ]',
				self.addRowChoice, multiline="command",
				update = self.setRPreview)
		    elif k == 'c':
			self.initEphEdit('New column choice: [ ^O: newline ]',
				self.addColChoice, multiline="command",
				update = self.setCPreview)
		    elif row >= 0 and ( k == 'R' or ( k == 'enter' and col == -1 ) ):
			self.initEphEdit('Row choice: [ ^O: newline ]',
				lambda h: self.editRowChoice(row, history=h),
				initialText = node.rowChoices[row].history,
				update = self.setRPreview,
				multiline="command")
		    elif col >= 0 and ( k == 'C' or ( k == 'enter' and row == -1 ) ):
			self.initEphEdit('Column choice: [ ^O: newline ]',
				lambda h: self.editColChoice(col, history=h),
				initialText = node.colChoices[col].history,
				update = self.setCPreview,
				multiline="command")
		    elif row >= 0 and k == 'meta r':
			self.initEphEdit('Row note: [ ^O: newline ]',
				lambda n: self.editRowChoice(row, note=n),
				initialText = node.rowChoices[row].note,
				multiline="command")
		    elif col >= 0 and k == 'meta c':
			self.initEphEdit('Column note: [ ^O: newline ]',
				lambda n: self.editColChoice(col, note=n),
				initialText = node.colChoices[col].note,
				multiline="command")
		    elif row >= 0 and ( k == 'meta R' or ( k == 'delete' and col == -1 ) ):
			self.confirmAction("Delete this row?",
				lambda: node.deleteRowChoice(row))
		    elif col >= 0 and ( k == 'meta C' or ( k == 'delete' and row == -1 ) ):
			self.confirmAction("Delete this column?",
				lambda: node.deleteColChoice(col))
		    elif row == -1 and col >= 0 and k == 'D':
			node.dominateColChoice(col)
		    elif col == -1 and row >= 0 and k == 'D':
			node.dominateRowChoice(row)
		    elif k == 'v' or (
			    k == 'enter' and row == -1 and col == -1 ):
			self.initEphEdit('Value of node:',
				lambda v: self.currentNode.setValuation(v),
				initialText = node.valuation.valuation)
		    elif k == 'e':
			node.evaluate()
		    elif k == 'M':
			node.clearMixed()
		    elif k == 'p':
			self.plotNode()
		    elif k == 'P':
			self.plotNode(True)
		    elif row >= 0 and col >= 0:
			child = node.children[row][col]
			if k == 'V':
			    self.initEphEdit('Value of child:',
				    lambda v:
				    self.currentNode.setChildValuation(
					row, col, v),
				    initialText = child.valuation.valuation)
			elif k == 'enter':
			    self.setCurrentNode(child)

		self.setPreviews()

	    elif self.inputState == 'upperEdit':
		k = self.editedWidget.keypress(size[0:1], k)
		if k in ['esc', 'ctrl g']:
		    self.inputState = 'top'
		    self.focusOnMatrix()
		    self.messageWidget.set_text('')
		    if self.editedWidget == self.historyWidgetWrap:
			self.currentNode.historyR = self.historyRWidget.get_edit_text()
			self.currentNode.historyC = self.historyCWidget.get_edit_text()
		    elif self.editedWidget == self.stateWidgetWrap:
			self.currentNode.state = self.stateWidget.get_edit_text()
		    elif self.editedWidget == self.notesWidgetWrap:
			self.currentNode.notes = self.notesWidget.get_edit_text()

	    elif self.inputState == 'ephEdit':
		k = self.ephEditWidget.keypress((size[0],), k)
		if k in ['esc', 'enter', 'ctrl g']:
		    text = self.ephEditWidget.get_edit_text()

		    self.messageWidget.set_text('')
		    self.inputState = 'top'
		    self.focusOnMatrix()
		    self.ephEditWidget.set_edit_text('')

		    if k == 'enter' and self.ephEditCallback:
			try: self.ephEditCallback(text)
			except UserError, err:
			    self.errorMessage(err.text)
		    self.setPreviews()
		elif self.ephEditUpdate:
		    self.ephEditUpdate(self.ephEditWidget.get_edit_text())

	    elif self.inputState == 'overlay':
		#k = self.top.keypress(size, k)
		k = self.helpTextWidgetWrapper.keypress(size, k)
		if k in ['esc', 'ctrl g']:
		    self.unOverlay()
	except UserError, err:
	    self.errorMessage(err.text)

    def overlay(self, top_widget):
	self.top = urwid.Overlay(top_widget, self.top, 'center', ('relative',
	    90), 'middle', ('relative', 80), 10, 2)
	self.inputState = 'overlay'
    def unOverlay(self):
	self.top = self.top.bottom_w
	self.inputState = 'top'

    def initEphEdit(self, caption, callback, multiline=False, initialText=None,
	    update=None, get_completions=None):
	self.messageWidget.set_text(caption)
	if initialText is None: initialText = ""
	self.ephEditWidget.set_edit_text(initialText)
	self.ephEditWidget.set_edit_pos(len(initialText))
	self.ephEditWidget.multiline=multiline
	if get_completions is not None:
	    self.ephEditWidget.allow_tab = "complete"
	    self.ephEditWidget.get_completions = get_completions
	    self.ephEditWidget.complete_list = None
	else:
	    self.ephEditWidget.allow_tab = True
	self.ephEditCallback = callback
	self.ephEditUpdate = update
	self.top.set_focus(self.ephEditWidget)
	self.inputState = 'ephEdit'

    def initUpperEdit(self, widget, message):
	self.selectUpperWidget(widget)
	self.inputState = 'upperEdit'
	self.editedWidget = widget
	self.messageWidget.set_text(message)
	self.top.set_focus(self.upper)

    ## some ephEdit callbacks:
    def addRowChoice(self, choiceHistory):
	if choiceHistory == '':
	    raise UserError("Bad label: empty!")
	self.currentNode.addRowChoice(Choice(choiceHistory))
    def addColChoice(self, choiceHistory):
	if choiceHistory == '':
	    raise UserError("Bad label: empty!")
	self.currentNode.addColChoice(Choice(choiceHistory))
    def editRowChoice(self, row, history=None, note=None):
	if history == '':
	    raise UserError("Bad label: empty!")
	self.currentNode.editRowChoice(row, history, note)
    def editColChoice(self, col, history=None, note=None):
	if history == '':
	    raise UserError("Bad label: empty!")
	self.currentNode.editColChoice(col, history, note)
	
    def setRPreview(self, choice):
	self.historyRPreviewWidget.set_text(rotateText(choice))
    def setCPreview(self, choice):
	self.historyCPreviewWidget.set_text(rotateText(choice))
    def setPreviews(self):
	row = self.matrixWidget.selectedRow
	col = self.matrixWidget.selectedCol
	node = self.currentNode
	if row >= 0: self.setRPreview(node.rowChoices[row].history)
	else: self.setRPreview('')
	if col >= 0: self.setCPreview(node.colChoices[col].history)
	else: self.setCPreview('')

    def checkpath(self, filename):
	if filename == '':
	    raise UserError("Empty file name")
	if chrooted:
	    # Simplistic method of preventing file access above the current
	    # directory.
	    # XXX: this assumes that the filename is only going to be passed
	    # directly to open(). Allowing anything else gets to mung it
	    # first may introduce security holes.
	    if filename[0] == '/' or filename.find('..') != -1:
		raise UserError("Restricted to this directory")
    def saveTree(self, filename):
	self.checkpath(filename)
	try:
	    if os.path.exists(filename):
		self.confirmAction(
			'"%s" exists: confirm overwrite' % filename,
			lambda: self.doSaveTree(filename))
	    else:
		self.doSaveTree(filename)
	except IOError, e:
	    raise UserError("%s: %s", (e.strerror, e.filename))
    def doSaveTree(self, filename):
	try:
	    #open(filename, 'w').write(self.currentNode.getRoot().flattenTree())
	    open(filename, 'w').write(self.currentNode.getRoot().pyTreeStr())
	    self.currentFilename = filename
	except IOError, e:
	    raise UserError("%s: %s", (e.strerror, e.filename))
    def loadTree(self, filename):
	#raise UserError("Cannot load - the appropriate module is temporarily "
	#"inaccessible, due to foolishness")
	self.checkpath(filename)
	try:
	    #self.setCurrentNode(unflattenTree(open(filename).read()))
	    self.setCurrentNode(readPyTree(open(filename).read()))
	    self.currentFilename = filename
	except IOError, e:
	    raise UserError("%s: %s" % (e.strerror, e.filename))
    def warlocksTree(self, str):
	try:
	    num = int(str)
	except:
	    num = snarfRavenblack.latestNum(str)
	self.setCurrentNode(snarfRavenblack.nodeOfNum(num))

    def errorMessage(self, message):
	self.messageWidget.set_text(('error',message))
	self.inputState = 'waitKey'
	def callback(k):
	    self.messageWidget.set_text('')
	    return k
	self.waitKeyCallback = callback

    def infoMessage(self, message):
	self.messageWidget.set_text(('new-info', message))
	self.inputState = 'waitKey'
	def callback(k):
	    self.messageWidget.set_text(('old-info', message))
	    return k
	self.waitKeyCallback = callback

    def confirmAction(self, message, action):
	self.messageWidget.set_text(('confirm', message + ' [y/N]'))
	self.inputState = 'waitKey'
	def callback(k):
	    self.messageWidget.set_text('')
	    if k in ['y','Y']: action()
	self.waitKeyCallback = callback

    def setCurrentNode(self, newNode):
	self.currentNode = newNode
	self.setWidgetContents()
	self.setPreviews()

    def resetMatrix(self):
	self.matrixWidget = self.currentNode.getMatrixWidget(
		self.matrixWidget.startRow, self.matrixWidget.startCol,
		self.matrixWidget.selectedRow, self.matrixWidget.selectedCol)
	self.matrixWidgetWrap.w = self.matrixWidget

    def draw_screen(self, size):
	canvas = self.top.render(size, focus=True)
	self.ui.draw_screen(size, canvas)
	
    def plotNode(self, forCol=False):
	if web_display.is_web_request():
	    raise UserError("Plotting not supported by web interface")

	try:
	    self.gnuplotPipe.sendCommands(self.currentNode.plot(forCol))
	except UserError, err:
	    raise err
	except:
	    raise
	    #raise UserError("Plotting failed - %s: %s" % (sys.exc_info()[0:2]))

    def quit(self):
	raise QuitException()

def main():
    urwid.web_display.set_preferences("SMGE")
    # try to handle short web requests quickly
    if urwid.web_display.handle_short_request():
	return

    if chDir:
	os.chdir(chDir)

    if web_display.is_web_request():
	import random
	w = random.choice(['exdeath','slartucker','zugzwang','toyotami',
		    'spacca', 'xade','freesoul','calicojack','citanest','reds'])
	try: interface = Interface(snarfRavenblack.nodeOfNum(
	    snarfRavenblack.latestNum(w)))
	except: interface = Interface(Node("nameless"))
	
    elif len(sys.argv) > 1:
	try:
	    interface = Interface(readPyTree(open(sys.argv[1]).read()),
		    sys.argv[1])
	except IOError, e:
	    interface = Interface(Node("nameless"), sys.argv[1])
    else:
	interface = Interface(Node("nameless", historyR=" \n\n\n ", historyC=" \n\n\n "))

    interface.infoMessage("Welcome to SMGE: type '?' for help")
    interface.main()

if __name__ == "__main__" or urwid.web_display.is_web_request():
    main()
